import pandas as pd

MONTHS = ["JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"]
YEARLY_FREQS = ["A", "BA", "AS", "BAS"]
QUARTERLY_FREQS = ["Q", "BQ", "QS", "BQS"]

VALID_YEARLY = YEARLY_FREQS + [f"{freq}-{month}" for freq in YEARLY_FREQS for month in MONTHS]
VALID_QUARTERLY = QUARTERLY_FREQS + [
    f"{freq}-{month}" for freq in QUARTERLY_FREQS for month in MONTHS
]
VALID_MONTHLY = ["M", "MS", "BM", "BMS"]

MONTHS_IN_YEAR = 12
MONTHS_IN_QUARTER = 3
QUARTERS_IN_YEAR = 4

FREQ_CONVERSION_FACTORS = {
    "yearly": {"monthly": MONTHS_IN_YEAR, "quarterly": QUARTERS_IN_YEAR},
    "quarterly": {"monthly": MONTHS_IN_QUARTER},
}

OFFSET_CONVERSIONS = {
    "yearly": {
        "monthly": {"months": MONTHS_IN_YEAR - 1},
        "quarterly": {"months": (QUARTERS_IN_YEAR - 1) * MONTHS_IN_QUARTER},
    },
    "quarterly": {"monthly": {"months": MONTHS_IN_QUARTER - 1}},
}

LONG_FREQ_TO_CODE = {"yearly": "A", "quarterly": "Q", "monthly": "M"}
CODE_TO_LONG_FREQ = {v: k for k, v in LONG_FREQ_TO_CODE.items()}

FREQ_TO_ORDER = {"yearly": 10, "quarterly": 9, "monthly": 8}

ORDER_TO_FREQ = {v: k for k, v in FREQ_TO_ORDER.items()}


def is_annual_freq(freq_str):
    return freq_str in VALID_YEARLY


def is_quarterly_freq(freq_str):
    return freq_str in VALID_QUARTERLY


def is_monthly_freq(freq_str):
    return freq_str in VALID_MONTHLY


VALIDATE_FUNCS = [is_annual_freq, is_quarterly_freq, is_monthly_freq]


def validate_freqs(*freqs):
    for freq in freqs:
        if not any([f(freq) for f in VALIDATE_FUNCS]):
            raise NotImplementedError(
                f"Only annual, quarterly and monthly frequencies are supported, found {freq}"
            )


def get_frequency_name(freq):
    if is_annual_freq(freq):
        return "yearly"
    if is_quarterly_freq(freq):
        return "quarterly"
    if is_monthly_freq(freq):
        return "monthly"

    return ""


def auto_step_down_base_freq(freq):
    if not isinstance(freq, str):
        freq = freq.name

    freq_name = get_frequency_name(freq)
    one_freq_down = FREQ_TO_ORDER[freq_name] - 1
    high_freq_name = ORDER_TO_FREQ.get(one_freq_down)

    if not high_freq_name:
        raise NotImplementedError(f"No frequency lower than {freq_name} currently supported")

    low_freq_code = LONG_FREQ_TO_CODE[freq_name]
    high_freq_code = LONG_FREQ_TO_CODE[high_freq_name]

    base, suffix = freq.split("-")
    new_base = base.replace(low_freq_code, high_freq_code)
    if high_freq_name not in ["yearly", "quarterly"]:
        return new_base

    return new_base + "-" + suffix


def get_high_freq_endpoints(
    endpoint_date, low_freq_name, high_freq_name, target_freq, mode="start"
):
    freq_base = target_freq.split("-")[0]

    # If it's a start-of-month business calendar, it can end up that the high_freq_start is not the correct date
    # e.g. if the 1st of the low_freq_start month happens to be a weekend. Need to shift it back to the true
    # first non-weekend date
    if mode == "start":
        offset = pd.DateOffset(**OFFSET_CONVERSIONS[low_freq_name][high_freq_name])
        high_freq_start = endpoint_date - offset

        if "S" in freq_base and "B" in freq_base:
            start_day = high_freq_start.day
            start_date = high_freq_start.isocalendar().weekday

            if start_day != 1:
                # Case 1: The 1st of the month is not a weekend. We go to the first.
                if (start_date - start_day + 1) not in [0, 7]:
                    high_freq_start -= pd.DateOffset(days=start_day - 1)

                # Case 2: The 1st of the month is a weekend. Shift to the first monday.
                else:
                    shift = min(max(start_day - 1, 1), 6)
                    high_freq_start -= pd.DateOffset(days=shift)

        # If we're dealing with a business calendar, check that the offset result is not on the weekend
        # and shift back to Friday if so.
        elif "B" in freq_base:
            start_day = high_freq_start.isocalendar().weekday
            days_to_friday = max(0, start_day - 5)
            high_freq_start -= pd.DateOffset(days=days_to_friday)
        return high_freq_start

    if mode == "end":
        offset = pd.DateOffset(**OFFSET_CONVERSIONS[low_freq_name][high_freq_name])
        high_freq_end = endpoint_date + offset
        if "S" in freq_base and "B" in freq_base:
            end_day = high_freq_end.day
            end_date = high_freq_end.isocalendar().weekday

            if end_day != 1:
                if (end_date - end_date - 1) not in [0, 7]:
                    high_freq_end += pd.DateOffset(days=end_day + 1)
                else:
                    shift = max(min(end_day + 1, 1), 6)
                    high_freq_end += pd.DateOffset(days=shift)
        return high_freq_end


def get_last_day(month, year):
    if month == 2:
        if year % 4 == 0:
            return 29
        return 28
    if month in [4, 6, 9, 11]:
        return 30

    return 31


def business_cal_adjust(date, adjust_forward=True):
    day = date.day
    weekday = date.weekday()

    if not adjust_forward:
        if day != 1:
            # Case 1: The 1st of the month is not a weekend. We go to the first.
            if (weekday - day + 1) not in [0, 6]:
                date -= pd.DateOffset(days=day - 1)

            # Case 2: The 1st of the month is a weekend. Shift to the first monday.
            else:
                shift = min(max(day - 1, 1), 5)
                date -= pd.DateOffset(days=shift)

    else:
        month = date.month
        year = date.year

        last_day = get_last_day(month, year)
        delta = last_day - day
        last_weekday = (weekday + delta) % 7
        if day != last_day:
            # Case 1: The last of the month is not a weekend. Go to the last.
            if last_weekday not in [5, 6]:
                date += pd.DateOffset(days=delta)

            # Case 2: The last day of the month is a weekend. Go to the last friday.
            else:
                shift = last_weekday - 4
                date += pd.DateOffset(days=delta - shift)

    return date


def get_frequency_names(df, target_freq):
    low_freq = df.index.freq or df.index.inferred_freq

    low_freq_name = get_frequency_name(low_freq)
    high_freq_name = get_frequency_name(target_freq)

    return low_freq_name, high_freq_name


def make_names_from_frequencies(df, target_freq):
    low_freq_name, high_freq_name = get_frequency_names(df, target_freq)

    var_name = ""
    if isinstance(df, pd.Series):
        var_name = df.name or "data"
    elif isinstance(df, pd.DataFrame):
        var_name = df.columns[0]

    return var_name, f"{low_freq_name}_{var_name}", f"{high_freq_name}_{var_name}"


def make_companion_index(df, target_freq):
    low_freq = df.index.freq or df.index.inferred_freq
    low_freq_name, high_freq_name = get_frequency_names(df, target_freq)

    if not FREQ_TO_ORDER[low_freq_name] > FREQ_TO_ORDER[high_freq_name]:
        raise ValueError(
            f"target_freq must be of higher frequency than the frequency on the data. Found "
            f"target_freq {target_freq}, which is {high_freq_name}, while data is {low_freq.name}, "
            f"which is {low_freq_name}."
        )

    low_freq_df = df.copy()
    start_date, end_date = low_freq_df.index[[0, -1]]

    offset = pd.DateOffset(**OFFSET_CONVERSIONS[low_freq_name][high_freq_name])
    base = target_freq.split("-")[0]

    if "S" in base:
        end_date += offset
    else:
        start_date -= offset

    if "B" in base:
        start_date = business_cal_adjust(start_date, adjust_forward=False)
        end_date = business_cal_adjust(end_date, adjust_forward=True)

    high_freq_index = pd.date_range(start=start_date, end=end_date, freq=target_freq)

    high_freq_index.freq = target_freq
    return high_freq_index


def align_and_merge_dfs(low_freq_df, high_freq_df):
    low_freq = low_freq_df.index.inferred_freq

    # these are stored as adverbs (e.g. yearly), so remove the -ly suffix
    attr = get_frequency_name(low_freq)[:-2]

    low_freq_idx = getattr(low_freq_df.index, attr)
    high_freq_idx = getattr(high_freq_df.index, attr)
    low_freq_set = set(low_freq_idx)
    high_freq_set = set(high_freq_idx)

    full_set = sorted(list(low_freq_set.union(high_freq_set)))
    C_mask = [x in low_freq_idx for x in full_set]

    df = pd.merge(low_freq_df, high_freq_df, left_index=True, right_index=True, how="outer")

    return df, C_mask



import unittest
from typing import Callable, List, Tuple

import numpy as np
import pandas as pd
from hypothesis import given
from hypothesis.strategies import SearchStrategy, composite, integers

from tsdisagg.time_conversion import (
    MONTHS,
    auto_step_down_base_freq,
    get_frequency_names,
    make_companion_index,
)


@composite
def freq(
    draw: Callable[[SearchStrategy[int]], int], base: str, suffix_list: List[str]
) -> Tuple[str, str, bool]:

    bases = [f"B{base}", f"{base}S", f"B{base}S"]
    suffixes = [f"-{x}" for x in suffix_list] + [""]

    n_bases = len(bases) - 1
    n_suffixes = len(suffixes) - 1

    base_idx = draw(integers(min_value=0, max_value=n_bases))
    suffix_idx = draw(integers(min_value=0, max_value=n_suffixes))
    to_df = bool(draw(integers(0, 1)))

    return bases[base_idx], suffixes[suffix_idx], to_df


class TestCompanionIndex(unittest.TestCase):
    @given(freq(base="A", suffix_list=MONTHS))
    def test_yearly_dataframe_to_monthly(self, params):
        base, suffix, test_df = params
        freq = base + suffix

        df = pd.Series(1, index=pd.date_range("1900-01-01", "1902-01-01", freq=freq), name="test")
        if test_df:
            df = df.to_frame()

        T = df.shape[0]

        target_freq = base.replace("A", "M")
        index = make_companion_index(df, target_freq)
        low_freq_name, high_freq_name = get_frequency_names(df, target_freq)

        self.assertEqual(target_freq, index.freq)

        high_freq_df = pd.Series(1, index=index, name=high_freq_name)
        result = pd.merge(df, high_freq_df, left_index=True, right_index=True, how="outer")

        self.assertEqual(result.shape[0], T * 12)

        block_matrix = result.values[:, 0].reshape(T, 12)
        self.assertEqual(np.all(np.isnan(block_matrix).sum(axis=1) == 11), True)

    @given(freq(base="A", suffix_list=MONTHS))
    def test_yearly_to_quarterly(self, params):
        base, suffix, test_df = params
        freq = base + suffix

        df = pd.Series(1, index=pd.date_range("1900-01-01", "1902-01-01", freq=freq), name="test")
        if test_df:
            df = df.to_frame()

        T = df.shape[0]

        target_freq = base.replace("A", "Q") + suffix
        index = make_companion_index(df, target_freq)
        self.assertEqual(target_freq, index.freq)

        low_freq_name, high_freq_name = get_frequency_names(df, target_freq)

        high_freq_df = pd.Series(1, index=index, name=high_freq_name)
        result = pd.merge(df, high_freq_df, left_index=True, right_index=True, how="outer")
        self.assertEqual(result.shape[0], T * 4)

        block_matrix = result.values[:, 0].reshape(T, 4)
        self.assertEqual(np.all(np.isnan(block_matrix).sum(axis=1) == 3), True)

    @given(freq(base="Q", suffix_list=MONTHS))
    def test_quarterly_to_monthly(self, params):
        base, suffix, test_df = params
        freq = base + suffix

        df = pd.Series(1, index=pd.date_range("1900-01-01", "1902-01-01", freq=freq), name="test")
        if test_df:
            df = df.to_frame()

        T = df.shape[0]
        target_freq = base.replace("Q", "M")
        index = make_companion_index(df, target_freq)
        self.assertEqual(target_freq, index.freq)

        low_freq_name, high_freq_name = get_frequency_names(df, target_freq)

        high_freq_df = pd.Series(1, index=index, name=high_freq_name)
        result = pd.merge(df, high_freq_df, left_index=True, right_index=True, how="outer")

        self.assertEqual(result.shape[0], T * 3)

        block_matrix = result.values[:, 0].reshape(T, 3)
        self.assertEqual(np.all(np.isnan(block_matrix).sum(axis=1) == 2), True)


class TestUtilities(unittest.TestCase):
    def test_step_down_frequency(self):
        freq = "AS-JAN"

        auto_freq = auto_step_down_base_freq(freq)
        self.assertEqual(auto_freq, "QS-JAN")

        freq = "BQ-MAR"
        auto_freq = auto_step_down_base_freq(freq)
        self.assertEqual(auto_freq, "BM")

        freq = "MS"
        self.assertRaises(NotImplementedError, auto_step_down_base_freq, freq)


if __name__ == "__main__":
    unittest.main()



import calendar
import unittest
from typing import Callable, List, Tuple

import pandas as pd
from hypothesis import given
from hypothesis.strategies import SearchStrategy, composite, integers

from tsdisagg.time_conversion import MONTHS, get_frequency_names, make_companion_index
from tsdisagg.ts_disagg import prepare_input_dataframes


@composite
def freq(
    draw: Callable[[SearchStrategy[int]], int], base: str, suffix_list: List[str]
) -> Tuple[str, str, str]:

    bases = [f"B{base}", f"{base}S", f"B{base}S"]
    suffixes = [f"-{x}" for x in suffix_list] + [""]

    n_bases = len(bases) - 1
    n_suffixes = len(suffixes) - 1

    base_idx = draw(integers(min_value=0, max_value=n_bases))
    suffix_idx = draw(integers(min_value=0, max_value=n_suffixes))

    year = draw(integers(min_value=1900, max_value=2000))
    month = draw(integers(min_value=1, max_value=12))
    day = draw(integers(min_value=1, max_value=calendar.monthrange(year, month)[1]))

    start_date = f"{year}-{month}-{day}"

    base_freq = bases[base_idx]
    suffix = suffixes[suffix_idx]

    return base_freq, suffix, start_date


class TestPandasIndex(unittest.TestCase):
    @given(freq(base="A", suffix_list=MONTHS))
    def test_dataframe_merge(self, params):
        base, suffix, start_date = params
        freq = base + suffix
        target_freq = base.replace("A", "Q") + suffix

        low_freq_df = pd.Series(
            1, index=pd.date_range(start_date, freq=freq, periods=20), name="test"
        )
        index = make_companion_index(low_freq_df, target_freq)
        low_freq_name, high_freq_name = get_frequency_names(low_freq_df, target_freq)

        high_freq_df = pd.Series(1, index=index, name=high_freq_name)
        result = pd.merge(low_freq_df, high_freq_df, left_index=True, right_index=True, how="outer")

        df, _, _ = prepare_input_dataframes(low_freq_df, None, target_freq, "denton")
        self.assertEqual(df.shape[0], result.shape[0])

    @given(freq(base="A", suffix_list=MONTHS))
    def test_dataframe_merge_A_to_M(self, params):
        base, suffix, start_date = params
        freq = base + suffix
        target_freq = base.replace("A", "M")

        low_freq_df = pd.Series(
            1, index=pd.date_range(start_date, freq=freq, periods=20), name="test"
        )
        index = make_companion_index(low_freq_df, target_freq)
        low_freq_name, high_freq_name = get_frequency_names(low_freq_df, target_freq)

        high_freq_df = pd.Series(1, index=index, name=high_freq_name)
        result = pd.merge(low_freq_df, high_freq_df, left_index=True, right_index=True, how="outer")

        df, _, _ = prepare_input_dataframes(low_freq_df, None, target_freq, "denton")

        self.assertEqual(df.shape[0], result.shape[0])


if __name__ == "__main__":
    unittest.main()



import unittest

import pandas as pd
import pandas.testing as pd_testing

from tsdisagg import disaggregate_series


class DisaggregationTests(unittest.TestCase):
    def assertDataframeEqual(self, a, b, msg):
        try:
            pd_testing.assert_frame_equal(a, b)
        except AssertionError as e:
            raise self.failureException(msg) from e

    def setUp(self):
        self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)
        self.exports_m = pd.read_csv("tests/data/exports_m.csv", index_col=0)
        self.exports_m.index = pd.date_range(
            start="1972-01-01", freq="MS", periods=self.exports_m.shape[0]
        )
        self.exports_m.columns = ["exports"]

        self.sales_a = pd.read_csv("tests/data/sales_a.csv", index_col=0)
        self.sales_a.index = pd.date_range(
            start="1975-01-01", freq="YS", periods=self.sales_a.shape[0]
        )
        self.sales_a.columns = ["sales"]

        self.exports_q = pd.read_csv("tests/data/exports_q.csv", index_col=0)
        self.exports_q.index = pd.date_range(
            start="1972-01-01", freq="QS-OCT", periods=self.exports_q.shape[0]
        )
        self.exports_q.columns = ["exports"]

        self.imports_q = pd.read_csv("tests/data/imports_q.csv", index_col=0)
        self.imports_q.index = pd.date_range(
            start="1972-01-01", freq="QS-OCT", periods=self.exports_q.shape[0]
        )
        self.imports_q.columns = ["imports"]

    def test_chow_lin(self):
        expected = pd.read_csv("tests/data/R_output_chow_lin.csv", index_col=0)
        expected.index = self.exports_q.index
        expected.columns = ["sales"]

        sales_q_chow_lin = disaggregate_series(
            self.sales_a,
            self.exports_q.assign(constant=1),
            method="chow-lin",
            agg_func="sum",
            optimizer_kwargs={"method": "powell"},
            verbose=False,
        )

        self.assertEqual(expected, sales_q_chow_lin.to_frame())

    def test_chow_lin_two_indicator(self):
        expected = pd.read_csv("tests/data/R_output_chow_lin_two_indicator.csv", index_col=0)
        expected.index = self.exports_q.index
        expected.columns = ["sales"]

        df2 = self.exports_q.merge(self.imports_q, left_index=True, right_index=True)

        sales_q_chow_lin = disaggregate_series(
            self.sales_a,
            df2.resample("QS-OCT").first().assign(constant=1),
            method="chow-lin",
            agg_func="sum",
            optimizer_kwargs={"method": "l-bfgs-b"},
            verbose=True,
        )

        self.assertEqual(expected, sales_q_chow_lin.to_frame())

    def test_denton(self):
        expected = pd.read_csv("tests/data/R_output_denton.csv", index_col=0)
        expected.index = pd.date_range(start="1975-01-01", freq="QS-OCT", periods=144)
        expected.columns = ["sales"]

        sales_q_denton = disaggregate_series(
            self.sales_a, method="denton", agg_func="sum", optimizer_kwargs={"method": "powell"}
        )
        self.assertEqual(expected, sales_q_denton.to_frame())

    def test_denton_cholette_w_constant(self):
        expected = pd.read_csv("tests/data/R_output_denton_cholette.csv", index_col=0)
        expected.index = pd.date_range(start="1975-01-01", freq="QS-OCT", periods=144)
        expected.columns = ["sales"]

        sales_q_dc = disaggregate_series(
            self.sales_a,
            method="denton-cholette",
            agg_func="sum",
            optimizer_kwargs={"method": "powell"},
        )

        self.assertEqual(expected, sales_q_dc.to_frame())

    def test_denton_cholette_w_indicator(self):
        expected = pd.read_csv("tests/data/R_output_denton_cholette_w_indicator.csv", index_col=0)
        expected.index = self.exports_q.index
        expected.columns = ["sales"]

        sales_q_dc = disaggregate_series(
            self.sales_a,
            high_freq_df=self.exports_q,
            method="denton-cholette",
            agg_func="sum",
            optimizer_kwargs={"method": "powell"},
            verbose=False,
        )

        self.assertEqual(expected, sales_q_dc.to_frame())

    def test_litterman_A_to_M(self):
        expected = pd.read_csv("tests/data/R_output_litterman_A_to_M.csv", index_col=0)
        expected.index = self.exports_m.index
        expected.columns = ["sales"]

        sales_m_litterman = disaggregate_series(
            self.sales_a,
            high_freq_df=self.exports_m.assign(Constant=1),
            method="litterman",
            agg_func="sum",
            optimizer_kwargs={"method": "nelder-mead"},
            verbose=False,
        )

        self.assertEqual(expected, sales_m_litterman.to_frame())


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
from scipy import linalg, stats
from scipy.optimize import minimize

from tsdisagg.time_conversion import (
    FREQ_CONVERSION_FACTORS,
    align_and_merge_dfs,
    auto_step_down_base_freq,
    get_frequency_name,
    make_companion_index,
    make_names_from_frequencies,
    validate_freqs,
)


def build_conversion_matrix(n, nl, i_len, C_mask=None, agg_func="sum"):
    excess = n - i_len * nl
    extra_periods = int(np.ceil(excess / i_len))

    C_mask = C_mask or np.full(nl, True)

    if agg_func == "sum":
        i = np.ones((i_len, 1))
    elif agg_func == "mean":
        i = np.ones((i_len, 1)) / i_len
    elif agg_func == "first":
        i = np.zeros((i_len, 1))
        i[0] = 1
    elif agg_func == "last":
        i = np.zeros((i_len, 1))
        i[-1] = 1

    C = linalg.kron(np.eye(nl + extra_periods), i.T)
    C = C[C_mask, :n]

    return C


def log_likelihood(nl, CΣCT, ul):
    sign, log_det = np.linalg.slogdet(CΣCT)

    return -nl / 2 * np.log(2 * np.pi) - 0.5 * (log_det + ul.T @ np.linalg.solve(CΣCT, ul))


def build_difference_matrix(n, h=0):
    Δ = np.eye(n)
    Δ[np.where(np.eye(n, k=-1))] = -1
    return np.linalg.matrix_power(Δ, h)


def build_distribution_matrix(Σ, C):
    return np.linalg.solve(C @ Σ @ C.T, C @ Σ).T


def build_chao_lin_covariance(rho, sigma_e_sq, n):
    row = rho ** np.arange(n)
    Σ_CL = np.r_[[np.r_[np.zeros(n - i), row[:i]] for i in range(n, 0, -1)]]
    Σ_CL += Σ_CL.T - np.eye(n)
    Σ_CL *= sigma_e_sq / (1 - rho)
    return Σ_CL


def build_litterman_covariance(rho, sigma_e_sq, n):
    Δ = build_difference_matrix(n, h=1)
    H_rho = np.eye(n, k=-1) * -rho + np.eye(n)
    Σ_L = sigma_e_sq * np.linalg.solve(Δ.T @ H_rho.T @ H_rho @ Δ, np.eye(n))
    return Σ_L


def GLS_beta_hat(Σ, y, X, C, nl):
    CΣCT_inv = np.linalg.solve(C @ Σ @ C.T, np.eye(nl))
    A = X.T @ C.T @ CΣCT_inv @ C @ X
    B = X.T @ C.T @ CΣCT_inv @ y
    β = np.linalg.solve(A, B)

    return β


def f_minimize(params, y, X, C, f_cov):
    n, k = X.shape
    nl = y.shape[0]

    ρ, sigma_e_sq = params
    Σ = f_cov(ρ, sigma_e_sq, n)
    β = GLS_beta_hat(Σ, y, X, C, nl)

    p = X @ β
    ul = y - C @ p
    CΣCT = C @ Σ @ C.T
    return -log_likelihood(nl, CΣCT, ul)


def build_denton_covariance(n, C, X, h=1, criterion="proportional"):
    Δ = build_difference_matrix(n, h)
    if criterion == "proportional":
        Δ = Δ @ np.diag(1 / X.ravel() / X.mean())
    Σ_D = np.linalg.solve(Δ.T @ Δ, np.eye(n))

    return Σ_D


def build_denton_charlotte_distribution_matrix(n, nl, C, X, h=1, criterion="proportional"):
    # Here is the Charlotte correction: slice off the top h rows of the difference matrix
    Δ = build_difference_matrix(n, h)[h:, :]
    if criterion == "proportional":
        Δ = Δ @ np.diag(1 / X.ravel() / X.mean())
    W_1 = np.r_[np.c_[Δ.T @ Δ, C.T], np.c_[C, np.zeros((nl, nl))]]
    W_2 = np.r_[np.c_[Δ.T @ Δ, np.zeros((n, nl))], np.c_[C, np.eye(nl)]]
    W = np.linalg.solve(W_1, W_2)

    w_theta = W[:n, n:]
    w_gamma = W[n:, n:]

    return w_theta


def print_regression_report(y, X, params, std_β, C, method):
    print(f"Dependent Variable: {y.name}")
    print(f"GLS Estimates using {method.title()}'s covariance matrix")
    print(f"N = {X.shape[0]}\t\tdf = {X.shape[0] - len(params)}")
    N, k = X.shape
    deg_f = N - len(params)
    t_dist = stats.t(df=deg_f)

    ul = y.dropna().values - C @ X.values @ params[:-2]
    r2 = 1 - np.var(ul) / (y.dropna() - y.mean()).var()
    adj_r2 = 1 - (1 - r2) * (N - 1) / (N - k - 1)

    print(f"Adj r2 = {adj_r2:0.4f}")
    print("")

    print(
        f'{"Variable":<15}{"coef":>10}{"sd err":>15}{"t":>15}{"P > |t|":>15}{"[0.025":>15}{"0.975]":>15}'
    )
    print("-" * 100)
    for i, var in enumerate(X.columns):
        t_05 = t_dist.ppf(1 - 0.05 / 2)
        ci_low = params[i] - std_β[i] * t_05
        ci_high = params[i] + std_β[i] * t_05

        t_stat = params[i] / std_β[i]
        p_value = t_dist.sf(np.abs(t_stat))
        print(
            f"{var:<15}{params[i]:>10.4f}{std_β[i]:>15.4f}{t_stat:>15.4f}{p_value:>15.4f}{ci_low:>15.4f}{ci_high:>15.4f}"
        )
    print("")
    print(f'{"rho":<15}{params[-2]:>10.4f}')
    print(f'{"sigma.sq":<15}{params[-1]:>10.4f}')


def prepare_input_dataframes(df1, df2, target_freq, method):
    df1_out = df1.copy()

    if not isinstance(df1.index, pd.core.indexes.datetimes.DatetimeIndex):
        raise ValueError(
            "No datetime index found on the dataframe passed as argument to low_freq_df."
        )

    if df1.isna().any().any():
        raise ValueError("low_freq_df has missing values.")

    if df2 is not None:
        if not isinstance(df2.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise ValueError(
                "No datetime index found on the dataframe passed as argument to low_freq_df."
            )

        if df2.isna().any().any():
            raise ValueError("high_freq_df has missing values.")

        if df2.index[0] > df1.index[0]:
            raise ValueError(
                f"Start date found on high frequency data {df2.index[0]} is after start date found on "
                f"low frequency data {df1.index[0]}. Interpolation is not possible in this case, because "
                f"there is no observed high frequency data associated with the first "
                f"{(df1.index < df2.index[0]).sum()} low-frequency observations. "
                f"Align the start date of these two dataframes and try again."
            )

        df2_out = df2.copy()
    else:
        df2_out = df2

    low_freq = df1_out.index.freq or df1_out.index.inferred_freq
    if not low_freq:
        raise ValueError(
            "Low frequency dataframe does not have a valid time index with frequency information"
        )

    if df2_out is None and target_freq is None:
        high_freq = auto_step_down_base_freq(low_freq)
    elif df2_out is None and target_freq is not None:
        high_freq = target_freq
    elif df2_out is not None and target_freq is not None:
        if df2_out.index.inferred_freq != target_freq:
            raise ValueError(
                "User provided target_freq does not match frequency information found on indicator data "
                "high_freq_df."
            )
        high_freq = target_freq
    else:
        high_freq = df2_out.index.inferred_freq
        if not high_freq:
            raise ValueError(
                "Indicator data high_freq_df does not have a valid time index with frequency information"
            )

    validate_freqs(low_freq, high_freq)

    high_name = get_frequency_name(high_freq)
    low_name = get_frequency_name(low_freq)
    time_conversion_factor = FREQ_CONVERSION_FACTORS[low_name][high_name]

    var_name, low_freq_name, high_freq_name = make_names_from_frequencies(df1_out, high_freq)

    if isinstance(df1_out, pd.Series):
        df1_out.name = low_freq_name
    elif isinstance(df1_out, pd.DataFrame):
        df1_out.rename(columns={var_name: low_freq_name}, inplace=True)

    if df2_out is None and method in ["denton", "denton-cholette"]:
        high_freq_idx = make_companion_index(df1_out, target_freq=high_freq)
        df2_out = pd.Series(1, index=high_freq_idx, name=high_freq_name)

    elif df2_out is None:
        raise ValueError(
            'high_freq_df can only be None for methods "denton" and "denton-cholette", otherwise a '
            "dataframe of high-frequency indicators must be provided."
        )
    df, C_mask = align_and_merge_dfs(df1_out, df2_out)

    return df, C_mask, time_conversion_factor


def disaggregate_series(
    low_freq_df,
    high_freq_df=None,
    target_freq=None,
    target_column=None,
    agg_func="sum",
    method="denton-cholette",
    criterion="proportional",
    h=1,
    optimizer_kwargs=None,
    verbose=True,
):

    """
    Transform a low frequency time series into a higher frequency series, preserving certain statistics aggregate
    statistics.

    Parameters
    ----------
    low_freq_df: DataFrame
        Low frequency time series to be converted
    high_freq_df: Dataframe, Optional
        High frequency companion data. The user can optionally supply data that is believed to be correlated with the
        low frequency data. This data will be used to interpolate the missing high frequency observations.
    target_freq: str, Optional
        The desired output frequency for the low frequency data. If None, the frequency of the low frequency data will
        be "stepped down" once, e.g. yearly to quarterly, or quarterly monthly. Should be a valid pandas frequency
        string; see the pandas documentation for details.
    target_column: str, default "None"
        If low_freq_df has multiple columns, the name of the column to be converted. If None, the first column in the
        dataframe will be used.
    agg_func: str, default "sum"
        Function which the interpolated high frequency data should respect when aggregating back to the original low
        frequency. One of: sum, mean, first, last.
    criterion: str, default "proportional"
        Objective function used by methods "denton" and "denton-cholette" to minimize the sum of square deviations from
        the low frequency data and the interpolated high frequency data. One of "proportional" (deviation) or
        "additive" (absolute deviation).
    method: str, default "denton"
        Method of interpolation to use. Currently, "denton", "denton-cholette", "chow-lin", and "litterman" are
        supported.
            Denton: Univariate method, naive interpolation that preserves the statistic specificed in the agg_func
                    argument only. The first several observation in the produced series will have unacceptably high
                    variance, so "denton-cholette" is recommended in all cases.
            Denton-Cholette: As Denton, but with an h-differences correction applied to reduce variance in the first
                    few observations.
            Chow-Lin: Multivariate regression-based method which allows for multiple high-frequency indicator series
                    to be provided. For details, see [1]
            Litterman: Multivariate regression-based method which allows for multiple high-frequency indicator series
                    to be provided. For details, see [1].
    h: int, default 1
        Number of differences to use when applying the Cholette correction to the Denton method. The Cholette method
        corrects for high variance in the first few observations when using the Denton method. A higher h will result
        in a more aggressive correction.
    optimizer_kwargs: dict, Optional
        A dictionary of keyword arguments to be passed to scipy.optimize.minimize. For full details, consult the
        docstring of that function. Ignoired if method is "denton" or "denton-cholette", as these do not use numerical
        optimization.
    verbose: bool, default True
        Whether to print regression results. Ignored if method is "denton" or "denton-cholette", as these have no
        regression results to report.

    Returns
    -------
    high_freq_df: Series
        A Pandas Series object containing the interpolated high frequency data.
    """

    if isinstance(low_freq_df, pd.Series):
        low_freq_df = low_freq_df.to_frame()

    if method not in ["denton", "denton-cholette", "chow-lin", "litterman"]:
        raise ValueError(
            f"Method should be one of 'denton', 'denton-cholette', 'chow-lin', 'litterman'. Got {method}."
        )
    if criterion not in ["proportional", "additive"]:
        raise ValueError(f"Criterion should be one of 'proportional', 'additive'. Got {criterion}")
    if agg_func not in ["mean", "sum", "first", "last"]:
        raise ValueError(
            f"agg_func should be one of 'mean', 'sum', 'first', 'last'. Got {agg_func}"
        )

    target_column = target_column or low_freq_df.columns[0]
    target_idx = np.flatnonzero(low_freq_df.columns == target_column)[0]

    df, C_mask, time_conversion_factor = prepare_input_dataframes(
        low_freq_df, high_freq_df, target_freq, method
    )

    y = df.iloc[:, target_idx].dropna().values
    X = df.drop(columns=df.columns[target_idx]).values

    n, k = X.shape
    nl = y.shape[0]

    C = build_conversion_matrix(n, nl, time_conversion_factor, agg_func=agg_func, C_mask=C_mask)
    if method == "denton":
        assert k == 1
        Σ = build_denton_covariance(n, C, X, h, criterion)
        D = build_distribution_matrix(Σ, C)
        p = X.ravel()

    elif method == "denton-cholette":
        assert k == 1
        D = build_denton_charlotte_distribution_matrix(n, nl, C, X, h, criterion)
        p = X.ravel()

    else:
        if optimizer_kwargs is None:
            optimizer_kwargs = {"method": "nelder-mead"}
        if optimizer_kwargs and "method" not in optimizer_kwargs.keys():
            optimizer_kwargs.update({"method": "nelder-mead"})

        if method == "chow-lin":
            f_cov = build_chao_lin_covariance
        elif method == "litterman":
            f_cov = build_litterman_covariance
        else:
            raise ValueError(f"Method {method} not supported.")

        # betas are unbounded, bound rho between 0 and 1 and sigma between 0 and +inf
        bounds = [(1e-5, 1 - 1e-5), (1e-5, None)]

        x0 = np.full(2, 0.8)
        result = minimize(
            f_minimize, x0=x0, args=(y, X, C, f_cov), bounds=bounds, **optimizer_kwargs
        )

        ρ, sigma_e_sq = result.x
        Σ = f_cov(ρ, sigma_e_sq, n)
        Σ_inv = np.linalg.solve(Σ, np.eye(n))

        β = GLS_beta_hat(Σ, y, X, C, nl)
        std_β = np.sqrt(np.diagonal(np.linalg.solve(X.T @ Σ_inv @ X, np.eye(k))))

        if verbose:
            print_regression_report(
                df.iloc[:, 0], df.iloc[:, 1:], np.r_[β, ρ, sigma_e_sq], std_β, C, method
            )

        p = X @ β
        D = build_distribution_matrix(Σ, C)

    ul = y - C @ p
    y_hat = p + D @ ul

    output = pd.Series(y_hat, index=df.index, name=target_column)
    output.index.freq = output.index.inferred_freq
    return output
